import { SessionCreationError } from "../../common/errors/SessionCreationError";
import { SessionNotFoundError } from "../../common/errors/SessionNotFoundError";
import { whileNotUndefined } from "../../common/utilities";
import { NetworkPlayer } from "../../game/entities/player/NetworkPlayer";
import { MatchSession } from "../../game/MatchSession";
import { PlayerNotificationService } from "./PlayerNotificationService";
import { matchSessionAdapter } from "../db/DbAdapter";
import { DbListResponse, DbMatchSession, DbMatchSessionUpdate } from "../db/interfaces";
import { MatchSessionMongoManager } from "../db/mongo/MatchSessionMongoManager";
import { SessionManager } from "../managers/SessionManager";
import { ServiceBase } from "./ServiceBase";
import { SocketIoService } from "./SocketIoService";
import { MatchSessionOptions } from "../../game/dto/MatchSessionOptions";

export class SessionService extends ServiceBase{
  private dbManager: MatchSessionMongoManager = new MatchSessionMongoManager();
  private sessionManager: SessionManager = new SessionManager();
  private notifyService = new PlayerNotificationService();  
  constructor() {
    super()
  }
  
  public async createSession(user: any, options: MatchSessionOptions ): Promise<string> {
    let socketClient;
    try {
      socketClient = await whileNotUndefined(() => SocketIoService.getClient(user._id));
    } catch (error) {
      throw new SessionCreationError();
    }
    const player =  new NetworkPlayer(user._id, user.username, socketClient.socketId);
    const session = new MatchSession(player, options);
    const dbSessionId = await this.dbManager.create(matchSessionAdapter(session));
    if (dbSessionId === undefined) {
      throw new SessionCreationError();
    }
    session.id = dbSessionId.toString();
    socketClient.sessionId = session.id;
    this.sessionManager.setSession(session);    
    this.notifyService.notifyMatchState(session);
    this.notifyService.notifyPlayersState(session.players);
    this.logger.debug(`Session ${session.id} created`);
    return session.id;
  }

  public async joinSession(user: any, sessionId: string): Promise<string> {
    const session: MatchSession | undefined = this.sessionManager.getSession(sessionId);
    if (session === undefined) {
      throw new SessionNotFoundError();
    }    
    let socketClient;
    try {
      socketClient = await whileNotUndefined(() => SocketIoService.getClient(user._id));
    } catch (error) {
      throw new SessionCreationError();
    }
    const player =  new NetworkPlayer(user._id, user.username, socketClient.socketId);
    this.dbManager.replaceOne({id: session.id}, matchSessionAdapter(session));
    session.addPlayerToSession(player);
    socketClient.sessionId = session.id;    
    this.notifyService.notifyMatchState(session);
    this.notifyService.notifyPlayersState(session.players);
    return sessionId
  }

  public async listJoinableSessions(): Promise<DbListResponse> {
    return await this.dbManager.listByFilter(
      { status: 'created' }, 
      { createdAt: -1 }, 
      { page: 1, pageSize: 12 }) as DbListResponse;
  }

  public async getSession(sessionId: string): Promise<DbMatchSession | undefined> {
    return await this.dbManager.getById(sessionId) as DbMatchSession | undefined;
  }

  public async deleteSession(sessionId: string): Promise<any> {
    this.sessionManager.deleteSession(sessionId);
    const session = {
      _id: sessionId,
      status: 'deleted'
    } as DbMatchSessionUpdate;
    return this.dbManager.update(session);
  }

  public updateSession(session: MatchSession): any {
    const entity = matchSessionAdapter(session);
    entity._id = session.id
    this.dbManager.update(entity);
  }
}