import { Server as HttpServer } from "http";
import { ServiceBase } from "./ServiceBase";
import { Server } from "socket.io";
import { SecurityManager } from "../managers/SecurityManager";
import { Socket } from "socket.io";
import { InteractionService } from "./InteractionService";
import { ClientEvents } from "../../game/constants";

export class SocketIoService  extends ServiceBase {
  io: Server
  private static clients: Map<string, any> = new Map();
  private interactionService: InteractionService = new InteractionService();

  static getClient(id: string) {
    return this.clients.get(id);
  }

  security = new SecurityManager();
  constructor(private httpServer: HttpServer) {
    super()
    this.io = this.socketIo(httpServer);  
    this.io.use(async (socket: Socket, next: any) => {
      if (socket.handshake.auth && socket.handshake.auth.token) {
        const token = socket.handshake.auth.token;
        try {          
          const { user } = await this.security.verifyJwt(token);
          socket.user = user;
          next();
        } catch (err) {
          this.logger.error(err);
          next(new Error('Authentication error'));
        }
      } else {
        next(new Error('Authentication error'));
      }
    }); 
    this.initListeners();
  }

  public getServer(): Server {
    return this.io;
  }

  getUserId(socket: Socket) {
    const { user } = socket;
    return user?._id?.toString();
  }

  private initListeners() {
    this.io.on('connection', (socket) => {
      this.logger.debug(`connect ${socket.id}`);
      if (socket.recovered) {
        // recovery was successful: socket.id, socket.rooms and socket.data were restored
        this.logger.debug("recovered!");
        this.logger.debug("socket.rooms:", socket.rooms);
        this.logger.debug("socket.data:", socket.data);
      } else {
        this.logger.debug("new connection");
        const { id: socketId, user } = socket;
        if (user !== undefined && user._id !== undefined) {
          const userId = user._id.toString();
          if (!SocketIoService.clients.has(userId)) {
            SocketIoService.clients.set(userId, { socketId, alive: true, user: socket.user  });
          } else {
            const client = SocketIoService.clients.get(userId);
            this.interactionService.updateSocketId(client.sessionId, userId, socketId);
            client.socketId = socketId;
            this.logger.debug(`User '${user.username}' already connected. Updating socketId to ${socketId}`);
            client.alive = true;
          }            
        } else {
          this.logger.error('User not found');
          socket.disconnect();
        }
      }    
      socket.on('disconnect', () => {
        const id = this.getUserId(socket);
        if (id) {
          this.logger.info('user disconnected');
          SocketIoService.clients.delete(id);
        }
      });

      socket.on('client:join-room', (room) => {
        socket.join(room);
        this.logger.debug(`User ${socket.user?.username} joined room ${room}`);
      });

      socket.on(ClientEvents.CLIENT_EVENT, (data) => {
        this.interactionService.handleClientEvent(data);
      });

      socket.on(ClientEvents.CLIENT_EVENT_WITH_ACK, (data, callback) => {
        const result = this.interactionService.handleClientEventWithAck(data);
        callback(result);
      });
      
      socket.on('pong', () => {
        const id = this.getUserId(socket);
        if (id && SocketIoService.clients.has(id)) {
          const client = SocketIoService.clients.get(id);
          SocketIoService.clients.set(id, {...client,  alive: true });
        }
      })
      socket.onAny((event, ...args) => {
        if (['pong'].includes(event)) return;
        let logStr = `Event received: ${event}`
        
        if (event.startsWith('client:') && args.length > 0) {
          logStr = `${logStr} (${args[0].event})`;
        }
        this.logger.trace(logStr);
      });
      
      this.pingClients()
    
      // // socket.on('createSession', (data, callback) => {
      // //   const response = sessionController.createSession(data, socket.id);
      // //   callback(response);
      // // });

      // socket.on('startSession', (data, callback) => {
      //   const response = this.sessionService.startSession(data);    
      //   callback(response);
      // });

      // socket.on('client:tile-animation-ended', (data) => {
      //   this.sessionService.onClientEndTileAnimation(data);
      // });
    
      //  socket.on('joinSession', (data, callback) => {
      //   const response = sessionController.joinSession(data, socket.id);
      //   callback(response);
      // });

      // socket.on('playerReady', (data, callback) => {
      //   const response = this.sessionService.setPlayerReady(data);
      //   callback(response);
      // });
    });    
  }

  private pingClients() {
    setInterval(() => {
      for (let [id, client] of SocketIoService.clients.entries()) {
        if (!client.alive) {
          this.logger.debug(`Client ${id} did not respond. Disconnecting.`);
          this.io.to(id).disconnectSockets(true); // Disconnect client
          SocketIoService.clients.delete(id);
        } else {
          client.alive = false; // Reset alive status for the next ping
          this.io.to(client.socketId).emit('ping'); // Send ping message
        }
      }
    }, 10000); 
  }
  
  private socketIo(httpServer: HttpServer): Server {
    return new Server(httpServer, {
      cors: {
        origin: '*',
      },
      connectionStateRecovery: {
        maxDisconnectionDuration: 15 * 60 * 1000,
        skipMiddlewares: true,
      },
      connectTimeout: 15 * 60 * 1000,
    })
  }
}   